#version 450
#define max_lights 100

layout (location = 0) out vec4 outColor;
//layout (location = 0) in vec4 fColor;
layout (location = 0) in vec3 position;
layout (location = 1) in vec3 normal_in;

layout (location = 2) in vec2 uv_in;

layout (binding = 0) uniform ub_helper_struct {
    mat4 model_matrix;
    mat4 view_matrix;
    mat4 projection_matrix;
    vec4 color;
    // ka, kd, ks, alpha
    vec4 material_properties;
} ub;

struct ub_directional_light_struct {
    vec3 direction;
    vec4 color;
};

struct ub_point_light_struct {
    vec3 position;
    vec4 color;
    // static, linear, quadratic
    vec3 attenuation;
};

struct ub_cone_light_struct {
    vec3 position;
    vec3 direction;
    float inner_cutoff;
    float outer_cutoff;
    vec4 color;
};

layout (binding = 1) uniform ub_lights_struct {
    int point_light_count;
    int directional_light_count;
    int cone_light_count;

    ub_point_light_struct point_light[max_lights];
    ub_directional_light_struct directional_light[1];
    ub_cone_light_struct cone_light[1];

    vec3 camera_position;
} ub_light;

layout (binding = 2) uniform sampler2D diffuse_texture;
layout (binding = 3) uniform sampler2D specular_texture;
layout (binding = 4) uniform samplerCube environment_texture;

void main() {
    vec3 position_in_world_space = (ub.model_matrix * vec4(position, 1)).xyz;

    float ka = ub.material_properties[0];
    float kd = ub.material_properties[1];
    float ks = ub.material_properties[2];
    float alpha = ub.material_properties[3];

//    vec4 Ia = vec4(1.0f, 1.0f, 1.0f, 1.0f);

    vec3 N = normalize(normal_in);
    // perspective_projection_matrix * view_transformation * var.view_transformations;

//    gl_Position = ub.projection_matrix * ub.view_matrix * ub.model_matrix * vec4(position, 1.0f);
    vec3 V = normalize(ub_light.camera_position - position_in_world_space);

//    vec3 Ip = ka * ub.color.xyz;
//    vec3 Id = ub.color.xyz;
    vec3 Ip = ka * texture(diffuse_texture, uv_in).rgb;
    vec3 Id = texture(diffuse_texture, uv_in).rgb;
    // Power of the reflection
    vec3 IsT = texture(specular_texture, uv_in).rgb;

    // where I is the vector pointing from the camera to the fragment's (interpolated) position
    // Basically my V
    vec3 reflect_to_cubemap = reflect(-V, N);
    vec3 environment_color = vec3(texture(environment_texture, reflect_to_cubemap));

    Ip += environment_color * IsT * 0.2;

    // * Point light
    for (int i = 0; i < 1; i++) {
        vec3 Lm = ub_light.point_light[i].position - position_in_world_space;
        float d = length(Lm);
        Lm = normalize(Lm);

        float attenutation = 1 / (
            ub_light.point_light[i].attenuation.x +
            ub_light.point_light[i].attenuation.y * d +
            ub_light.point_light[i].attenuation.z * d * d
        );

        //        float attenutation = 1;

        vec3 Rm = normalize(2 * dot(Lm, N) * N - Lm);

        // * 5.12
        vec3 Is = ub_light.point_light[i].color.xyz;

        Ip += (attenutation * (kd * max(dot(Lm, N), 0) * Id + (ks+1) * pow(max(dot(Rm, V), 0.0f), alpha) * Is * IsT));
//        Ip += vec3(environment_color);
//        Ip = theta * vec3(1, 1, 1);
    }

    // * Directional Light
    for (int i = 0; i < 1; i++) {
        ub_directional_light_struct this_light = ub_light.directional_light[i];
        vec3 Lm = -normalize(this_light.direction);
        vec3 Rm = normalize(2 * dot(Lm, N) * N - Lm);
        vec3 Is = this_light.color.xyz;
//        vec3 Is = texture(specular_texture, uv_in).rgb;

        Ip += (kd * max(dot(Lm, N), 0) * Id + ks * pow(max(dot(Rm, V), 0.0f), alpha) * Is * IsT);

        //        fColor = vec4(Lm, 1.0f);
        //        fColor = vec4(dot(Lm, N).xxx, 1.0f);
    }

    // * Spotlight
    for (int i = 0; i < 1; i++) {
        vec4 spec = vec4(0.0);
        ub_cone_light_struct this_light = ub_light.cone_light[i];

        vec3 Lm = normalize(this_light.position - position_in_world_space);
        vec3 Sd = normalize(vec3(this_light.direction));
        float smoothness = 0;

        // inside the cone?
        if (dot(Sd, -Lm) > this_light.outer_cutoff) {
            vec3 Rm = normalize(2 * dot(Lm, N) * N - Lm);
            vec3 Is = this_light.color.xyz;
//            vec3 Is = texture(specular_texture, uv_in).rgb;

            smoothness = 1 - smoothstep(this_light.inner_cutoff, this_light.outer_cutoff, dot(-Lm, Sd));
//            smoothness = 1;

            Ip += (kd * max(dot(Lm, N), 0) * Id + ks * pow(max(dot(Rm, V), 0.0f), alpha) * Is * IsT) * smoothness;
        }
    }

    outColor = vec4(Ip, 1.0f);
}
