#version 450
#define max_lights 100

layout (binding = 0) uniform ub_helper_struct {
    mat4 model_matrix;
    mat4 view_matrix;
    mat4 projection_matrix;
    vec4 color;
    // ka, kd, ks, alpha
    vec4 material_properties;
} ub;

layout (location = 0) in vec3 position_in;
layout (location = 0) out vec3 position_out;

layout (location = 1) in vec3 normal_in;
layout (location = 1) out vec3 normal_out;

layout (location = 2) in vec2 uv_in;
layout (location = 2) out vec2 uv_out;

struct ub_directional_light_struct {
    vec3 direction;
    vec4 color;
};

struct ub_point_light_struct {
    vec3 position;
    vec4 color;
// static, linear, quadratic
    vec3 attenuation;
};

struct ub_cone_light_struct {
    vec3 position;
    vec3 direction;
    float inner_cutoff;
    float outer_cutoff;
    vec4 color;
};

layout (binding = 1) uniform ub_lights_struct {
    int point_light_count;
    int directional_light_count;
    int cone_light_count;

    ub_point_light_struct point_light[max_lights];
    ub_directional_light_struct directional_light[1];
    ub_cone_light_struct cone_light[1];

    vec3 camera_position;
} ub_light;

void main() {
    gl_Position = ub.projection_matrix * ub.view_matrix * ub.model_matrix * vec4(position_in, 1.0f);
    position_out = position_in;

    normal_out = normal_in;
    uv_out = uv_in;
}